"""Messenger gRPC grpc-messenger-client."""
import dataclasses
from datetime import datetime
from typing import Dict, Generator, List, Optional, Sequence, Type, TypeVar, Union
from uuid import UUID

import grpc
from fred_api.messenger import service_email_grpc_pb2_grpc, service_letter_grpc_pb2_grpc, service_sms_grpc_pb2_grpc
from fred_api.messenger.common_types_pb2 import (Reference as MessengerReference,
                                                 ReferenceType as MessengerReferenceType, Uuid)
from fred_api.messenger.letter_pb2 import Address as MessengerAddress
from fred_api.messenger.service_email_grpc_pb2 import GetEmailRequest, ListEmailRequest, SendEmailRequest
from fred_api.messenger.service_letter_grpc_pb2 import GetLetterRequest, ListLetterRequest, SendLetterRequest
from fred_api.messenger.service_sms_grpc_pb2 import GetSmsRequest, ListSmsRequest, SendSmsRequest
from frgal import GrpcClient, GrpcDecoder
from google.protobuf.struct_pb2 import Struct
from typist import SecretaryClient

from .email import Email, EmailEnvelope
from .letter import Letter, LetterEnvelope
from .message import Message, MessageEnvelope, Reference, ReferenceType
from .sms import Sms, SmsEnvelope

E = TypeVar('E', bound=MessageEnvelope)


class HermesError(Exception):
    """Base class for Hermes related errors."""


class NoMessage(HermesError):
    """Raised when no message matching the UUID is found."""


class SendFailed(HermesError):
    """Raised when sending of the message failed."""


class MessengerDecoder(GrpcDecoder):
    """Messenger gRPC decoder."""

    def __init__(self):
        """Initialize the decoder object by registering custom decoders."""
        super().__init__()
        self.set_decoder(MessengerReference, self._decode_reference)
        self.set_decoder(Struct, self._decode_protobuf_struct)
        self.set_decoder(Uuid, self._decode_uuid)

    def _decode_reference(self, value: MessengerReference) -> Reference:
        """Decode Reference."""
        if value.type == MessengerReferenceType.registrar:
            return Reference(type=ReferenceType(MessengerReferenceType.Name(value.type)), value=value.value)
        else:
            return Reference(type=ReferenceType(MessengerReferenceType.Name(value.type)), value=UUID(value.value))

    def _decode_protobuf_struct(self, value: Struct) -> Dict:
        """Decode google.protobuf.Struct to Dict."""
        return {k: self.decode(v) for k, v in value.items()}

    def _decode_uuid(self, value: Uuid) -> Optional[UUID]:
        """Decode Uuid."""
        if not value.value:
            return None
        return UUID(value.value)


class MessengerClient(GrpcClient):
    """Messenger gRPC grpc-messenger-client."""

    email_service = "EmailMessenger"
    letter_service = "LetterMessenger"
    sms_service = "SmsMessenger"

    def __init__(
        self,
        netloc: str,
        credentials: grpc.ChannelCredentials = None,
        *,
        secretary_client: SecretaryClient = None
    ):
        """Initialize Logger GRPC backend grpc-messenger-client instance.

        Args:
            netloc: Network location of a gRPC server.
            credentials: Credentials for a secure channel connection. If None, insecure channel is used.
            secretary_client: Instance of django-secretary REST API client.

        """
        decoder = MessengerDecoder()
        super().__init__(
            netloc,
            decoder=decoder,
            grpc_modules=[service_email_grpc_pb2_grpc, service_letter_grpc_pb2_grpc, service_sms_grpc_pb2_grpc],
            credentials=credentials,
        )
        self.secretary_client = secretary_client

    def send_email(self, email: Email, *, archive: bool = True, references: Sequence[Reference] = None) -> str:
        """Send gRPC request to send email request.

        Args:
            email: Email structure to be sent
            archive: Whether the message should be archived
            references: List of registry object references.

        Raises:
            SendFailed: When sending of the message fails.
        """
        request = SendEmailRequest()
        request.message_data.recipient = email.recipient
        request.message_data.subject_template = email.subject_template
        request.message_data.body_template = email.body_template
        request.message_data.extra_headers.update(**(email.extra_headers or {}))
        request.message_data.context.update(email.context)
        if email.sender is not None:
            request.message_data.sender = email.sender
        if email.body_template_html is not None:
            request.message_data.body_template_html = email.body_template_html
        if email.subject_template_uuid is not None:
            request.message_data.subject_template_uuid.value = str(email.subject_template_uuid)
        if email.body_template_uuid is not None:
            request.message_data.body_template_uuid.value = str(email.body_template_uuid)
        if email.body_template_html_uuid is not None:
            request.message_data.body_template_html_uuid.value = str(email.body_template_html_uuid)
        for uuid in email.attachments or ():
            request.message_data.attachments.add(value=str(uuid))
        return self._send_message(
            grpc_service=self.email_service,
            request=request,
            message=email,
            archive=archive,
            references=references,
        )

    def get_email(self, uid: str) -> EmailEnvelope:
        """Send gRPC request to get email."""
        return self._get_message(
            grpc_service=self.email_service,
            request_cls=GetEmailRequest,
            envelope_cls=EmailEnvelope,
            uid=uid,
        )

    def list_emails(self, **kwargs) -> Generator[EmailEnvelope, None, None]:
        """Send gRPC request to get email list."""
        return self._list_messages(
            grpc_service=self.email_service,
            request_cls=ListEmailRequest,
            envelope_cls=EmailEnvelope,
            **kwargs,
        )

    def send_letter(self, letter: Letter, *, archive: bool = True, references: Sequence[Reference] = None) -> str:
        """Send gRPC request to send letter.

        Args:
            letter: Letter structure to be sent.
            archive: Whether the message should be archived.
            references: List of registry object references.

        Raises:
            SendFailed: When sending of the message fails.
        """
        request = SendLetterRequest()
        request.message_data.recipient.CopyFrom(MessengerAddress(**dataclasses.asdict(letter.recipient)))
        request.message_data.file.value = str(letter.file)
        return self._send_message(
            grpc_service=self.letter_service,
            request=request,
            message=letter,
            archive=archive,
            references=references,
        )

    def get_letter(self, uid: str) -> LetterEnvelope:
        """Send gRPC request to get letter."""
        return self._get_message(
            grpc_service=self.letter_service,
            request_cls=GetLetterRequest,
            envelope_cls=LetterEnvelope,
            uid=uid,
        )

    def list_letters(self, **kwargs) -> Generator[LetterEnvelope, None, None]:
        """Send gRPC request to get letter list."""
        return self._list_messages(
            grpc_service=self.letter_service,
            request_cls=ListLetterRequest,
            envelope_cls=LetterEnvelope,
            **kwargs,
        )

    def send_sms(self, sms: Sms, *, archive: bool = True, references: Sequence[Reference] = None) -> str:
        """Send gRPC request to send sms request.

        Args:
            sms: Sms structure to be sent
            archive: Whether the message should be archived
            references: List of registry object references.

        Raises:
            SendFailed: When sending of the message fails.
        """
        request = SendSmsRequest()
        request.message_data.recipient = sms.recipient
        request.message_data.body_template = sms.body_template
        request.message_data.context.update(sms.context)
        if sms.body_template_uuid is not None:
            request.message_data.body_template_uuid.value = str(sms.body_template_uuid)
        return self._send_message(
            grpc_service=self.sms_service,
            request=request,
            message=sms,
            archive=archive,
            references=references,
        )

    def get_sms(self, uid: str) -> SmsEnvelope:
        """Send gRPC request to get sms."""
        return self._get_message(
            grpc_service=self.sms_service,
            request_cls=GetSmsRequest,
            envelope_cls=SmsEnvelope,
            uid=uid,
        )

    def list_sms(self, **kwargs) -> Generator[SmsEnvelope, None, None]:
        """Send gRPC request to get sms list."""
        return self._list_messages(
            grpc_service=self.sms_service,
            request_cls=ListSmsRequest,
            envelope_cls=SmsEnvelope,
            **kwargs,
        )

    def _get_message(
        self,
        grpc_service: str,
        request_cls: Union[GetEmailRequest, GetLetterRequest, GetSmsRequest],
        envelope_cls: Type[E],
        uid: str
    ) -> E:
        """Send gRPC request to get message.

        Args:
            grpc_service: Name of grpc service implementing ``get`` method.
            request_cls: Get request message class.
            envelope_cls: Message envelope class.

            uid: Unique message identification.

        Raises:
            NoMessage: When no message matching uid is found.
        """
        request = request_cls()
        request.uid.value = uid
        try:
            message_data = self.call(grpc_service, "get", request)
        except grpc.RpcError as error:
            if error.code() == grpc.StatusCode.NOT_FOUND:
                raise NoMessage('Message {} not found'.format(uid))
            raise
        return envelope_cls.from_dict(message_data, secretary_client=self.secretary_client)

    def _send_message(
        self,
        grpc_service: str,
        request: Union[GetEmailRequest, GetLetterRequest, GetSmsRequest],
        message: Message,
        *,
        archive: bool = True,
        references: Sequence[Reference] = None
    ) -> str:
        """Send gRPC request to send message request.

        Args:
            grpc_service: Name of grpc service implementing ``get`` method.
            request: Send request message.
            message: Message structure to be sent
            archive: Whether the message should be archived
            references: List of registry object references.

        Raises:
            SendFailed: When sending of the message fails.
        """
        request.archive = archive
        if message.type:
            request.message_data.type = message.type
        if references:
            request.references.extend(self._get_messenger_references(references))
        try:
            return self.call(grpc_service, "send", request)
        except grpc.RpcError as error:
            raise SendFailed('Sending of the message failed: {}'.format(error))

    def _list_messages(
        self,
        grpc_service: str,
        request_cls: Union[ListEmailRequest, ListLetterRequest, ListSmsRequest],
        envelope_cls: Type[E],
        created_from: datetime = None,
        created_to: datetime = None,
        recipients: List[str] = None,
        body_templates: List[str] = None,
        types: List[str] = None,
        limit: int = None,
        references: List[Reference] = None
    ) -> Generator[E, None, None]:
        """Send gRPC request to get list of messages.

        Args:
            grpc_service: Name of grpc service implementing ``list`` method.
            request_cls: List request message class.
            envelope_cls: Message envelope class.

            created_from: Lower limit for create datetime.
            created_to: Upper limit for create datetime.
            recipients: List of recipients to filter.
            body_templates: List of body template names to filter.
            types: List of message types to filter.
            limit: Maximum number of messages returned.
            references: List of registry object references.
        """
        request = request_cls()
        if created_from:
            request.created_from.FromDatetime(created_from)
        if created_to:
            request.created_to.FromDatetime(created_to)
        if recipients:
            request.recipients[:] = recipients
        if body_templates:
            request.body_templates[:] = body_templates
        if types:
            request.types[:] = types
        if references:
            request.references.extend(self._get_messenger_references(references))
        if limit:
            request.limit = limit

        message_list = self.call_stream(grpc_service, "list", request)

        for message_data in message_list:
            yield envelope_cls.from_dict(message_data, secretary_client=self.secretary_client)

    def _get_messenger_references(self, references: Sequence[Reference]):
        messenger_references = []
        for reference in references:
            messenger_references.append(
                MessengerReference(
                    type=MessengerReferenceType.Value(reference.type.value),
                    value=str(reference.value),
                )
            )
        return messenger_references
